#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fluoroethene.h"

/* 外部変数 */
extern const Sword SwordList[SWORD_VARIATION];

/* 静的関数プロトタイプ宣言 */
static void next(int *self, int n, int dly);

/* 剣師の初期化 */
void ken_init(Kenshi *self, int hp, int atk, int dfs, int spd, const int *swd) {
	int i;

	self->max_hp = self->hp = hp;
	self->attack = atk;
	self->defense = dfs;
	self->speed = spd;
	for (i=0; swd[i]!=-1; i++) {
		self->swords[i] = swd[i];
		self->state[i] = SS_SLEEP;
	}
	self->swords[i] = -1;
	self->poison = 0;
	self->stun = 0;
	self->guard = 0;
	self->reflection = 0;
	self->life = 0;
        self->heavy_push = 0;
}

/* 剣師のコピー */
void ken_cpy(const Kenshi *src, Kenshi *dest) {
	ken_init(dest, src->hp, src->attack, src->defense, src->speed, src->swords);
}

/* マアグヱリテ表記化 */
void ken_to_theki(const Kenshi *self, char *dest) {
	int i;

	sprintf(dest, "%d/%d/%d/%d/", self->hp, self->attack, self->defense, self->speed);
	for (i=0; self->swords[i]!=-1; i++) {
		strcat(dest, SwordList[self->swords[i]].name);
	}
}

/* マアグヱリテ表記でファイルに追記 */
void ken_add_theki(const Kenshi *self, char *filename) {
	FILE *fp;
	char buff[64];

	ken_to_theki(self, buff);
	if ((fp=fopen(filename, "a")) != NULL){
		fprintf(fp, "%s\n", buff);
		fclose(fp);
	} else {
		fprintf(stderr, "file open failed at %s\n", buff);
	}
}

/* マアグヱリテ表記読み込み */
void ken_read_theki(Kenshi *self, const char *str){
	int i, j, prms[4], swd[SWORD_SIZE];
	char substr[8];

	for (i=0; i<4; i++) {
		for (j=0; *str!='/'; j++) {
			substr[j] = *str++;
		}
		substr[j] = '\0';
		str++;
		prms[i] = atoi(substr);
	}
	i = 0;
	while ((swd[i] = read_sword(&str)) >= 0)
          i++;
	swd[i] = -1;
	ken_init(self, prms[0], prms[1], prms[2], prms[3], swd);
}

/* APのセットと剣の使用済み解除・起動 */
void ken_set_ap(Kenshi *self, int turn){
	int i;

	self->ap = self->speed;
	for (i=0; self->swords[i]!=-1; i++) {
		if (self->state[i]!=SS_LOST && SwordList[self->swords[i]].awake<=turn) {
			self->state[i] = SS_ACTIVE;
		}
	}
}

/* あいざわの確認 */
int ken_aizawa(Kenshi *self){
	int i, s, allow;

	s = self->speed;
	/* すばやさ0で剣を持っていたらあいざわ */
	if (s==0 && self->swords[0]!=-1) {
		return 1;
	}
	/* まずは素で使える剣について調べる */
	allow = 0;
	for (i=0; i<s; i++) {
		/* 剣が尽きた場合は非あいざわ。加速剣を使えるなら判定が複雑なので非あいざわということにする */
		if (self->swords[i]==-1 || self->swords[i]==10) {
			return 0;
		/* 使用可能なリミテッド剣はあいざわとは関係のないものとする */
		} else if (SwordList[self->swords[i]].limited) {
			s++;
			continue;
		/* 残りは使用可能なアンリミ剣 */
		} else {
			if (allow < SwordList[self->swords[i]].awake) {
				allow = SwordList[self->swords[i]].awake;
			}
		}
	}
	/* 素で使える分より右にある剣について調べる */
	for (; self->swords[i]!=-1; i++) {
		/* 右の剣が左の剣より起動が遅ければ加速剣がない限りあいざわ */
		if (SwordList[self->swords[i]].awake >= allow) {
			for (; self->swords[i]!=-1; i++) {
				if (self->swords[i] == 10) {
					for (i--; i>=0; i--) {
						if (SwordList[self->swords[i]].awake == 1 && !SwordList[self->swords[i]].limited) {
							if (--s <= 0) {
								return 1; /* 加速剣が存在するが，その前に素早さ個以上の非遅延剣があるのであいざわ */
							}
						}
					}
					return 0; /* 加速剣が存在し，(高速剣の疑似遅延を考えなければ)明らかに使用可能なので非あいざわ */
				}
			}
			return 1;
		} else {
			/* 加速剣が使えるなら非あいざわということにする */
			if (self->swords[i]==10) {
				return 0;
			}
			/* ここに許容遅延更新則を入れたい */
		}
	}
	/* 全部の剣を使えるっぽいので非あいざわということにする */
	return 0;
}

/* オランダの確認 */
int ken_oranda(Kenshi *self){
	int i;

	if (self->attack != 0) { /* 攻撃力がある */
		for (i=0; self->swords[i]!=-1; i++) {
			/* 攻撃力依存の剣があるなら非オランダ */
			if (self->swords[i] <= 5) {
				return 0;
			}
		}
		/* ないのでオランダ */
		return 1;
	} else { /* 攻撃力がないのは非オランダとする */
		return 0;
	}
}

/* 次の剣構成 */
static void next(int *self, int n, int dly){
	int i;

	for (i=n-1; i>=0; i--) {
		self[i] += 1;
		if (self[i] == 14+dly) {
			self[i] = 0;
		} else {
			break;
		}
	}
}

/* 全ての簡易あいざわでも純正オランダでもない途心toshin剣師を引数としてprocを呼び出す */
void ken_each_in(int toshin, void (*proc)(Kenshi *self), int dly){
	int i, flg;
	int hp, attack, defense, speed, s;
	int swords[SWORD_SIZE];
	Kenshi x;

	/* HPが0の剣師 */
	for (attack=1; attack<=toshin-2; attack++) {
	for (defense=0; defense<=toshin-attack-2; defense++) {
		speed = toshin - attack - defense;
		if ((speed&1) == 0) {
			speed >>= 1;
			for (i=0; i<speed; i++) {
				swords[i] = 1;
			}
			swords[i] = -1;
			ken_init(&x, 0, attack, defense, speed, swords);
			proc(&x);
		}
	}}

	/* HPが0でない剣師 */
	for (hp=1; hp<=toshin; hp++) {
	for (attack=0; attack<=toshin-hp; attack++) {
	for (defense=0; defense<=toshin-hp-attack; defense++) {
	for (speed=0; speed<=toshin-hp-attack-defense; speed++) {
		s = toshin - hp - attack - defense - speed;
		if (speed==0 && s!=0) {
			continue;
		}
		for (i=0; i<s; i++) {
			swords[i] = 0;
		}
		swords[i] = -1;
		if (s == 0) {
			ken_init(&x, (hp<<2)+hp, attack, defense, speed, swords);
			proc(&x);
			continue;
		}
		for (;;) {
			ken_init(&x, (hp<<2)+hp, attack, defense, speed, swords);
			if (!(ken_aizawa(&x) || ken_oranda(&x))) {
				proc(&x);
			}
			next(swords, s, dly);
			flg = 1;
			for (i=0; i<s; i++) {
				if (swords[i]!=0) {
					flg = 0;
					break;
				}
			}
			if (flg) {
				break;
			}
		}
	}}}}
}
